import importlib
from typing import Dict, List, Optional

from pycosnippets.pyco import PyCo
from pycosnippets.utils import (
    json_pretty_print,
    json_save,
)
from pycosnippets.settings import (
    PYCO_TEST_SNIPPET,
)


class Main:

    def __init__(self):
        self.pyco = PyCo.instance()

    @json_pretty_print.decorator
    def hello_world(self, world: Optional[str] = None, template_path: Optional[str] = None) -> List[Dict]:
        return self.run_snippet(
            code=PYCO_TEST_SNIPPET.format(world=world or "Pythonista"),
            template_path=template_path,
        )

    @json_pretty_print.decorator
    @json_save.metadecorator(overwrite=True)
    def run_gist(
            self,
            user: str,
            gist_id: str,
            file: str,
            commit: Optional[str] = None,
            save: Optional[str] = None,
            template_path: Optional[str] = None
    ) -> List[Dict]:
        return self.pyco.run_gist(
            u=user,
            g=gist_id,
            f=file,
            c=commit,
            template_path=template_path
        )

    @json_pretty_print.decorator
    @json_save.metadecorator(overwrite=True)
    def run_script(self, filepath: str, save: Optional[str] = None, template_path: Optional[str] = None) -> List[Dict]:
        return self.pyco.run_script(
            filepath=filepath,
            template_path=template_path
        )

    @json_pretty_print.decorator
    @json_save.metadecorator(overwrite=True)
    def run_snippet(self, code: str, save: Optional[str] = None, template_path: Optional[str] = None) -> List[Dict]:
        return self.pyco.run_code(
            code_snippet=code,
            template_path=template_path
        )



import os
import sys
import time
from uuid import uuid4
from typing import Dict, List, Optional, Tuple
from multiprocessing import Process, Manager
from traitlets.config.configurable import SingletonConfigurable

import IPython

from pycosnippets.gists import Gist
from pycosnippets.templates import Template
from pycosnippets.utils import get_code_id
from pycosnippets.settings import PYCO_NOTEBOOK_TEMPLATE_OUTPUT_PATH


def run_code(snippet: str, template_path: Optional[str] = None) -> Tuple[str, List[Dict]]:
    template = Template(path=template_path)
    notebook = template.add_snippet(code=snippet, new_filepath=PYCO_NOTEBOOK_TEMPLATE_OUTPUT_PATH)
    notebook.execute()
    last_code_id = notebook.metadata["snippets"][-1]["code_id"]
    last_outputs = notebook.get_last_outputs()
    return last_code_id, last_outputs


def run_worker(execution_id: str, executions: Dict, snippet: str, template_path: Optional[str] = None):
    code_id, output = run_code(snippet=snippet, template_path=template_path)
    executions[execution_id] = {
        "code_id": code_id,
        "output": output
    }


class PyCo(SingletonConfigurable):

    def __init__(self):
        # TODO: delete all the multiprocessing implementation
        # TODO: Save both the executions & cache in a external datastore (i.e. Redis)
        # self.manager = Manager()
        self.executions = {}  # self.manager.dict()
        self.cache = {}  # TODO: load previous state

    def info(self) -> Dict:
        return {
            "pyco_id": id(self),
            "timestamp": time.time(),
            "records": {
                "executions": len(self.executions),
                "cache": len(self.cache)
            },
            "mem": {
                "executions": sys.getsizeof(self.executions),
                "cache": sys.getsizeof(self.cache)
            }
        }

    @property
    def ipython(self):
        return IPython.get_ipython()

    def magic(self, line: str):
        return self.ipython.magic(line)

    @staticmethod
    def get_gist(u: str, g: str, f: str, c: Optional[str] = None) -> str:
        gist = Gist(username=u, gist_id=g, filename=f, commit=c)
        return gist.get_code()

    def load_code(self, code_snippet: str):
        user_ns = getattr(self.ipython, 'user_ns', globals())
        exec(compile(code_snippet, f"{int(time.time())}", "exec"), user_ns, user_ns)

    def load_gist(self, u: str, g: str, f: str, c: Optional[str] = None):
        gist = Gist(username=u, gist_id=g, filename=f, commit=c)
        code_snippet = gist.get_code()
        self.load_code(code_snippet=code_snippet)

    @staticmethod
    def get_script(filepath: str) -> str:
        if not os.path.exists(filepath):
            raise ValueError(f"Script not found: {filepath}")
        with open(filepath, "r") as file:
            return file.read()

    def load_script(self, filepath: str):
        code_snippet = self.get_script(filepath)
        self.load_code(code_snippet=code_snippet)

    def run_code(
            self,
            code_snippet: str,
            template_path: Optional[str] = None,
            use_cache: bool = False,
            external_metadata: Optional[Dict] = None,
    ) -> Dict:
        if not use_cache:
            return self.run_code_without_cache(
                code_snippet=code_snippet,
                template_path=template_path,
                external_metadata=external_metadata
            )
        _, *code_id = get_code_id(code=code_snippet).split("-")
        output = self.cache.get("".join(code_id))
        return output if output else self.run_code(
            use_cache=False,
            code_snippet=code_snippet,
            template_path=template_path,
            external_metadata=external_metadata
        )

    def run_code_without_cache(
            self,
            code_snippet: str,
            template_path: Optional[str] = None,
            external_metadata: Optional[Dict] = None
    ) -> Dict:
        start = time.time()
        # Create an unique execution id to facilitate output retrieval
        execution_id = str(uuid4())
        run_worker(execution_id, self.executions, code_snippet, template_path)
        # Get the execution output and save to cache
        res = self.executions[execution_id]
        _, *code_id = res["code_id"].split("-")
        output = {
            "timestamp_start": start,
            "timestamp_final": time.time(),
            "output": res["output"],
            "metadata": external_metadata
        }
        self.cache["".join(code_id)] = output
        return output

    def run_gist(
            self,
            u: str,
            g: str,
            f: str,
            c: Optional[str] = None,
            template_path: Optional[str] = None,
            use_cache: bool = False,
    ) -> Dict:
        gist = Gist(username=u, gist_id=g, filename=f, commit=c)
        code_snippet = gist.get_code()
        return self.run_code(
            code_snippet=code_snippet,
            template_path=template_path,
            use_cache=use_cache,
        )

    def run_script(
            self,
            filepath: str,
            template_path: Optional[str] = None,
            use_cache: bool = False,
    ) -> Dict:
        code_snippet = self.get_script(filepath)
        return self.run_code(
            code_snippet=code_snippet,
            template_path=template_path,
            use_cache=use_cache,
        )



import os


PYCO_TIMEOUT = int(os.environ.get(
    "PYCO_TIMEOUT",
    default=600
))

PYCO_BASE_PATH = os.path.dirname(os.path.abspath(__file__))

# Jupyter Notebook

PYCO_NOTEBOOK_VERSION = int(os.environ.get(
    "PYCO_NOTEBOOK_VERSION",
    default=4
))

# Testing

PYCO_TEST_SNIPPET = """
pyco.load_gist(
    u="rhdzmota",
    g="9aa444dfb15b5d8a96d1204bf173309a",
    f="hello.py",
)

hello(world="{world}")
"""

# Templates settings

PYCO_NOTEBOOK_TEMPLATE_NAME = os.environ.get(
    "PYCO_NOTEBOOK_TEMPLATE",
    default="default.ipynb"
)

PYCO_NOTEBOOK_TEMPLATE_PATH = os.environ.get(
    "PYCO_NOTEBOOK_TEMPLATE_PATH",
    default=os.path.join(PYCO_BASE_PATH, "templates", PYCO_NOTEBOOK_TEMPLATE_NAME)
)

PYCO_NOTEBOOK_TEMPLATE_OUTPUT_PATH = os.environ.get(
    "PYCO_NOTEBOOK_TEMPLATE_OUTPUT_PATH",
    default=os.path.join(PYCO_BASE_PATH, ".out")
)

# Create template output path if not exists
os.makedirs(PYCO_NOTEBOOK_TEMPLATE_OUTPUT_PATH, exist_ok=True)

# Commandline

PYCO_CMD_OUTPUT_DIRECTORY = os.environ.get(
    "PYCO_CMD_OUTPUT_DIRECTORY",
    default="."
)



from typing import Optional

import requests


class Gist:
    URL_TEMPLATE = "https://gist.githubusercontent.com/{username}/{gist_id}/raw/{commit}/{filename}"

    def __init__(self, username: str, gist_id: str, filename: str, commit: Optional[str] = None):
        self.username = username
        self.gist_id = gist_id
        self.filename = filename
        self.commit = commit

    @property
    def url(self) -> str:
        return self.URL_TEMPLATE.format(
            username=self.username,
            gist_id=self.gist_id,
            filename=self.filename,
            commit=self.commit or ""
        )

    def get_code(self) -> str:
        response = requests.get(url=self.url)
        return response.text



from setuptools import setup, find_packages

with open("README.md") as file:
    readme = file.read()

with open("requirements.txt") as file:
    requirements = [req for req in file.read().split("\n") if req and not req.startswith("#")]


setup(
    name="pycosnippets",
    version="0.1.0",
    description="Python Code Snippets",
    long_description=readme,
    long_description_content_type='text/markdown',
    author="rhdzmota",
    author_email='contact@rhdzmota.com',
    url="https://github.com/rhdzmota/pycosnippets",
    classifiers=[
        'Programming Language :: Python :: 3.7',
    ],
    package_dir={
        "": "src"
    },
    package_data={
        "": [
            "templates/default.ipynb"
        ]
    },
    packages=find_packages(where='src'),
    include_package_data=True,
    scripts=[
        "bin/pycosnippets"
    ],
    install_requires=requirements,
    python_requires='>=3.7, <3.8',
    license="MIT",
)



import os
import time
import hashlib
import json
from typing import Dict, List, Optional

import nbformat
from nbconvert.preprocessors import ExecutePreprocessor, CellExecutionError

from pycosnippets.settings import (
    PYCO_NOTEBOOK_TEMPLATE_PATH,
    PYCO_NOTEBOOK_VERSION,
)


def validate(path: str) -> bool:
    return os.path.exists(path)


def get_id(code: str) -> str:
    now = time.time()
    return f"""{int(now)}-{hashlib.md5(code.encode("utf-8")).hexdigest()}"""


class Template:

    def __init__(
            self,
            path: Optional[str] = None,
            version: Optional[int] = None,
            allow_multiple_snippets: bool = False,
            snippet_lock: bool = False,
            metadata: Optional[Dict] = None
    ):
        self.path = path or PYCO_NOTEBOOK_TEMPLATE_PATH
        self.version = version or PYCO_NOTEBOOK_VERSION
        self.allow_multiple_snippets = allow_multiple_snippets
        self.snippet_lock = snippet_lock
        self.metadata = metadata or {}

    def read(self) -> Dict:
        with open(self.path, "r") as file:
            content = file.read()
        return json.loads(content)

    def write(
            self,
            data: Dict,
            new_filename: Optional[str] = None,
            new_filepath: Optional[str] = None,
    ) -> str:
        content = json.dumps(data, indent=2)
        filename = new_filename or os.path.basename(self.path)
        filepath = new_filepath or os.path.dirname(self.path)
        notebook_path = os.path.join(filepath, filename)
        with open(notebook_path, "w") as file:
            file.write(content)
        return notebook_path

    def get_last_outputs(self) -> List[Dict]:
        data = self.read()
        return data.get("cells", [{}])[-1].get("outputs", [])

    def add_snippet(
            self,
            code: str,
            new_filename: Optional[str] = None,
            new_filepath: Optional[str] = None,
            allow_multiple_snippets: bool = False,
    ) -> 'Template':
        if self.snippet_lock and not self.allow_multiple_snippets:
            raise ValueError("Existing snippet detected.")
        code_id = get_id(code)
        data = self.read()
        for cell in data.get("cells", []):
            if cell["cell_type"] != "code":
                continue
            source = cell.get("source", [])
            if not len(source):
                continue
            first_line, *_ = source
            if not first_line.startswith("# Code Snippet"):
                continue
            cell["source"] = code
            break
        path = self.write(
            data=data,
            new_filename=new_filename or f"{code_id}.ipynb",
            new_filepath=new_filepath
        )
        metadata_snippets = [*self.metadata.get("snippets", []), {"code_id": code_id, "code": code}]
        return Template(
            path=path,
            version=self.version,
            allow_multiple_snippets=allow_multiple_snippets,
            snippet_lock=True,
            metadata={
                **self.metadata,
                "snippets": metadata_snippets,
            }
        )

    def get_notebook_node(self) -> nbformat.notebooknode.NotebookNode:
        with open(self.path) as file:
            return nbformat.read(file, as_version=4)

    def execute(self):
        ep = ExecutePreprocessor(timeout=600)
        nb = self.get_notebook_node()
        try:
            dirpath = os.path.dirname(self.path)
            ep.preprocess(nb, {"metadata": {"path": dirpath}})
        except CellExecutionError:
            print(self.path)
            raise
        finally:
            with open(self.path, mode="w", encoding="utf-8") as file:
                nbformat.write(nb, file, version=4)



import os
import json
import time
import hashlib
import functools
from typing import Callable, Optional
from traitlets.config.configurable import SingletonConfigurable

from pycosnippets.settings import PYCO_CMD_OUTPUT_DIRECTORY


def get_code_id(code: str) -> str:
    now = time.time()
    return f"""{int(now)}-{hashlib.md5(code.encode("utf-8")).hexdigest()}"""


class JsonPrettyPrint(SingletonConfigurable):

    def __call__(self, obj: dict, **kwargs):
        string = json.dumps(obj, indent=4, **kwargs)
        print(string)

    def decorator(self, func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            output = func(*args, **kwargs)
            self(obj=output)
        return wrapper

    def metadecorator(self, **json_kwargs) -> Callable:
        def decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                output = func(*args, **kwargs)
                self(obj=output, **json_kwargs)
            return wrapper
        return decorator


json_pretty_print = JsonPrettyPrint.instance()


class JsonSave(SingletonConfigurable):

    def __call__(self, obj: dict, filepath: str, overwrite: bool = False, **kwargs) -> dict:
        string = json.dumps(obj, indent=4, **kwargs)
        if os.path.exists(filepath) and not overwrite:
            raise ValueError(f"File already exists and overwrite flag is set to False: {filepath}")
        with open(filepath, "w") as file:
            file.write(string)
        return obj

    def decorator(self, func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            save_path = kwargs.get("save")
            output = func(*args, **kwargs)
            if not save_path:
                return output
            self(obj=output, filepath=save_path)
        return wrapper

    def metadecorator(
            self,
            filepath: Optional[str] = None,
            overwrite: bool = False,
            save_path_argname: Optional[str] = None,
            **kwargs
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                save_path = kwargs.get(save_path_argname or "save", filepath)
                output = func(*args, **kwargs)
                if not save_path:
                    return output
                self(obj=output, filepath=save_path, overwrite=overwrite, **kwargs)
            return wrapper
        return decorator


json_save = JsonSave.instance()
